"""Define schemas for charge density data."""

from __future__ import annotations

from typing import TYPE_CHECKING
from warnings import warn

from pydantic import BaseModel, Field

from emmet.core.types.typing import DateTimeType, IdentifierType
from emmet.core.vasp.models import ChgcarLike

if TYPE_CHECKING:
    from typing import Any


class VolumetricDataDoc(BaseModel):
    """
    Volumetric data metadata for selected materials.
    """

    last_updated: DateTimeType = Field(
        description="Timestamp for the most recent update to the charge density data.",
    )

    task_id: IdentifierType | None = Field(
        None,
        description="The Materials Project ID of the calculation producing the charge density data. "
        "This comes in the form: mp-******.",
    )

    volumetric_file: str = Field(
        description="The name of the VASP file, e.g., CHGCAR, AECCAR0, etc."
    )

    volumetric_data: ChgcarLike = Field(
        description="The volumetric data from the CHGCAR-like VASP file."
    )


class ChgcarDataDoc(VolumetricDataDoc):
    """Legacy alias for backward compat."""

    def model_post_init(self, context: Any) -> None:
        """Raise deprecation warning."""
        warn(
            "ChgcarDataDoc is deprecated and will be removed in a "
            "future version of emmet-core. Please migrate to "
            "VolumetricDataDoc.",
            category=DeprecationWarning,
            stacklevel=2,
        )



"""Define basic schema for band theoretic properties."""

from __future__ import annotations

from collections import defaultdict
import numbers
from typing import TYPE_CHECKING, Annotated

import numpy as np
from pydantic import BaseModel, Field, computed_field, BeforeValidator
from pymatgen.core import Lattice
from pymatgen.electronic_structure.bandstructure import (
    BandStructure as PmgBandStructure,
)
from pymatgen.electronic_structure.bandstructure import Kpoint
from pymatgen.electronic_structure.core import Orbital, Spin
from pymatgen.electronic_structure.dos import CompleteDos, Dos

from emmet.core.math import Matrix3D, Vector3D
from emmet.core.types.pymatgen_types.structure_adapter import StructureType

if TYPE_CHECKING:
    from collections.abc import Callable
    from pymatgen.core.sites import PeriodicSite
    from typing_extensions import Self

BAND_GAP_TOL = 1e-4


class BandTheoryBase(BaseModel):

    identifier: str | None = Field(None, description="The identifier of this object.")
    structure: StructureType | None = Field(
        None, description="The structure associated with this calculation."
    )


def _deser_lattice(lattice: Lattice | dict | Matrix3D) -> Matrix3D:
    """Ensure the lattice matrix is stored only."""
    if isinstance(lattice, Lattice):
        return lattice.matrix
    elif isinstance(lattice, dict):
        return lattice.get("lattice")
    return lattice


class BandStructure(
    BandTheoryBase, populate_by_name=True, validate_by_alias=True, validate_by_name=True
):
    """Define the schema of band structures.

    This class is generic enough to accommodate both
    electronic and phonon band structures.
    """

    qpoints: list[Vector3D] = Field(
        description="The wave vectors (q-points) at which the band structure was sampled, in direct coordinates.",
    )

    reciprocal_lattice: Annotated[Matrix3D, BeforeValidator(_deser_lattice)] = Field(
        description="The reciprocal lattice.", validation_alias="lattice_rec"
    )

    labels_dict: dict[str, Vector3D] = Field(
        {}, description="The high-symmetry labels of specific q-points."
    )


class ProjectedBS(BaseModel):
    """Light memory model for projected band structure data.

    Only C order is used.

    When the data is unraveled by numpy, you should get a 4-index
    tensor where:
        - Index 1: band index
        - Index 2: k-point index
        - Index 3: site index
        - Index 4: index for the orbital character, ordered as
            s, py, pz, px, dxy, dyz, dz2, dxz, dx2-y2,...
    """

    spin_up: list[float] | None = Field(
        None, description="The flattened spin-up band projections."
    )
    spin_down: list[float] | None = Field(
        None, description="The flattened spin-down band projections."
    )
    rank: tuple[int, int, int, int] = Field(
        description="The original shape of the band projections."
    )

    @classmethod
    def from_pmg_like(cls, projections: dict[Spin, np.ndarray]) -> Self:

        spins = list(projections)
        rank = projections[spins[0]].shape
        if not all(proj.shape == rank for proj in projections.values()):
            raise ValueError(
                "The band structure projections should have the same shape."
            )
        config = {
            f"spin_{spin.name}": proj.flatten(order="C")
            for spin, proj in projections.items()
        }
        return cls(
            **config,  # type: ignore[arg-type]
            rank=rank,  # type: ignore[arg-type]
        )

    def to_pmg_like(self) -> dict[Spin, np.ndarray]:
        return {
            spin: np.array(getattr(self, f"spin_{spin.name}")).reshape(
                self.rank, order="C"
            )
            for spin in Spin
            if getattr(self, f"spin_{spin.name}")
        }


class ElectronicBS(BandStructure):
    """Define an electronic band structure schema."""

    efermi: float = Field(
        description="The Fermi level (highest occupied energy level.)"
    )

    band_gap: float | None = Field(None, description="The value of the band gap.")

    spin_up_bands: list[list[float]] | None = Field(
        None,
        description="The eigen-energies of the spin-up electrons. The first index represents the band, the second the k-point index.",
    )

    spin_down_bands: list[list[float]] | None = Field(
        None,
        description="The eigen-energies of the spin-down electrons. The first index represents the band, the second the k-point index.",
    )

    is_direct: bool | None = Field(
        None, description="If the bandgap is non-zero, whether the band gap is direct."
    )

    projections: ProjectedBS | None = Field(
        None, description="The site- and spin-orbital-projected band structure."
    )

    @computed_field
    def is_metal(self) -> bool:
        """Whether the band gap is almost zero."""
        return (
            self.band_gap < BAND_GAP_TOL
            if isinstance(self.band_gap, numbers.Number)
            else False
        )

    @classmethod
    def from_pmg(cls, ebs: PmgBandStructure, **kwargs) -> Self:
        """Construct from a pymatgen band structure object."""
        band_gap_meta = ebs.get_band_gap()
        config = {
            "qpoints": [qpt.frac_coords for qpt in ebs.kpoints],
            "lattice_rec": ebs.lattice_rec.matrix,
            "efermi": ebs.efermi,
            "labels_dict": {
                label: qpt.frac_coords for label, qpt in ebs.labels_dict.items()
            },
            "structure": ebs.structure,
            "is_metal": ebs.is_metal(),
            "is_direct": band_gap_meta["direct"],
            "band_gap": band_gap_meta["energy"],
        }

        for spin in Spin:
            config[f"spin_{spin.name}_bands"] = ebs.bands.get(spin)
        config["projections"] = (
            ProjectedBS.from_pmg_like(ebs.projections)  # type: ignore[arg-type]
            if ebs.projections
            else None
        )
        return cls(**config, **kwargs)

    def to_pmg(self, pmg_cls: Callable = PmgBandStructure) -> PmgBandStructure:
        """Construct the pymatgen object from the current instance.

        Parameters
        -----------
        pmg_cls : PmgBandStructure or a subclass
            Because BandStructureSymmLine has the same constructor
            signature as PmgBandStructure, any PmgBandStructure-derived
            class which has the same signature can be used here.
        """
        rlatt = Lattice(self.reciprocal_lattice)

        bands = {}
        for spin in Spin:
            if v := getattr(self, f"spin_{spin.name}_bands", None):
                bands[spin] = np.array(v)

        return pmg_cls(
            [Kpoint(q, lattice=rlatt).frac_coords for q in self.qpoints],  # type: ignore[misc]
            bands,
            rlatt,
            self.efermi,
            labels_dict={
                k: Kpoint(v, lattice=rlatt).frac_coords  # type: ignore[misc]
                for k, v in (self.labels_dict or {}).items()
            },
            coords_are_cartesian=False,
            structure=self.structure,
            projections=self.projections.to_pmg_like() if self.projections else None,
        )


class ProjectedDos(BaseModel):
    """Atom and orbital projected DOS."""

    spin_up_densities: list[list[float] | None] | None = Field(
        None, description="The spin-up projected densities of state."
    )

    spin_down_densities: list[list[float] | None] | None = Field(
        None, description="The spin-down projected densities of state."
    )

    orbital: list[str | None] | None = Field(
        None, description="The orbital character of this DOS, if applicable."
    )

    site_index: list[int | None] | None = Field(
        None,
        description="The index of the atom in the structure onto which this DOS is projected.",
    )

    def _to_list_of_dict(self) -> list[dict[str, str | int | list[float]]]:
        """Possible serialization procedure."""
        return [
            {k: getattr(self, k)[i] for k in self.__class__.model_fields}
            for i in range(len(self.spin_up_densities or []))
        ]

    @classmethod
    def _from_list_of_dict(cls, lod: list[dict[str, str | int | list[float]]]) -> Self:
        """Possible deserialization route."""
        return cls(**{k: [entry.get(k) for entry in lod] for k in cls.model_fields})  # type: ignore[arg-type]

    @classmethod
    def from_pmg_like(
        cls,
        pdos: dict[PeriodicSite, dict[Orbital, dict[Spin, np.ndarray]]],
        structure: StructureType,
    ) -> Self:
        """Create a ProjectedDos from a pymatgen-like CompleteDos.pdos."""
        projs = [
            {
                **{
                    f"spin_{spin.name}_densities": sr_dos
                    for spin, sr_dos in spin_dos.items()
                },
                "orbital": orbital.name,
                "site_index": [
                    idx for idx, ref_site in enumerate(structure) if ref_site == site
                ][0],
            }
            for site, orb_spin_dos in pdos.items()
            for orbital, spin_dos in orb_spin_dos.items()
        ]

        return cls(
            **{  # type: ignore[arg-type]
                k: [proj.get(k) for proj in projs]
                for k in (
                    "spin_up_densities",
                    "spin_down_densities",
                    "orbital",
                    "site_index",
                )
            },
        )

    def to_pmg_like(
        self, structure: StructureType
    ) -> dict[PeriodicSite, dict[Orbital, dict[Spin, np.ndarray]]]:
        """Construct a pymatgen-like representation of the projected DOS."""
        pdos: defaultdict[
            PeriodicSite, defaultdict[Orbital, dict[Spin, np.ndarray]]
        ] = defaultdict(lambda: defaultdict(dict))
        for i, dens in enumerate(self.spin_up_densities or []):
            pdos[site := structure[self.site_index[i]]][  # type: ignore[index]
                orb := Orbital[self.orbital[i]]  # type: ignore[index,misc]
            ][Spin.up] = np.array(dens)
            if self.spin_down_densities and all(
                sdd for sdd in self.spin_down_densities
            ):
                pdos[site][orb][Spin.down] = np.array(self.spin_down_densities[i])

        return dict(pdos)


class ElectronicDos(BandTheoryBase):
    """Electronic density of states (DOS) with possible projections."""

    spin_up_densities: list[float] | None = Field(
        None, description="The spin-up densities of state."
    )

    spin_down_densities: list[float] | None = Field(
        None, description="The spin-down densities of state."
    )

    energies: list[float] | None = Field(
        None, description="The energies at which the DOS was calculated."
    )

    efermi: float | None = Field(None, description="The Fermi energy.")

    projected_densities: ProjectedDos | None = Field(
        None, description="The orbital- and site-projected DOS."
    )

    @property
    def _available_spins(self) -> list[Spin]:
        """Retrive spin indices of non-null spin data."""
        return [spin for spin in Spin if getattr(self, f"spin_{spin.name}_densities")]

    @classmethod
    def from_pmg(cls, dos: Dos | CompleteDos, **kwargs) -> Self:
        """Create an electronic DOS from a pymatgen object."""

        densities: dict[str, list[float]] = {
            f"spin_{spin.name}_densities": sr_dos.tolist()
            for spin, sr_dos in dos.densities.items()
        }

        pdos = (
            ProjectedDos.from_pmg_like(dos.pdos, dos.structure)  # type: ignore[arg-type]
            if isinstance(dos, CompleteDos)
            else None
        )

        return cls(
            **densities,  # type: ignore[arg-type]
            efermi=dos.efermi,
            energies=dos.energies.tolist(),
            structure=getattr(dos, "structure", None),
            projected_densities=pdos,
            **kwargs,
        )

    def to_pmg(self) -> Dos | CompleteDos:
        """Serialize to pymatgen."""
        if self.efermi is None:
            raise ValueError(
                "Fermi level unspecified, cannot create a pymatgen (Complete)Dos object."
            )

        densities: dict[Spin, np.ndarray] = {
            spin: np.array(getattr(self, f"spin_{spin.name}_densities"))
            for spin in self._available_spins
        }

        dos = Dos(self.efermi, np.array(self.energies), densities)
        if self.structure and self.projected_densities:
            return CompleteDos(
                self.structure,
                dos,
                self.projected_densities.to_pmg_like(self.structure),
            )
        return dos



"""Core definition of a Provenance Document"""

from __future__ import annotations

import json
import warnings
from typing import TYPE_CHECKING, Any, Annotated

from pybtex.database import BibliographyData, parse_string
from pybtex.errors import set_strict_mode
from pydantic import BaseModel, Field, BeforeValidator
from pymatgen.core import Lattice, Structure, PeriodicSite

from emmet.core.material_property import PropertyDoc
from emmet.core.math import Matrix3D
from emmet.core.symmetry import SymmetryData
from emmet.core.types.enums import ValueEnum
from emmet.core.types.pymatgen_types.structure_adapter import StructureType
from emmet.core.types.pymatgen_types.lattice_adapter import LatticeType
from emmet.core.types.typing import DateTimeType, IdentifierType

if TYPE_CHECKING:
    from typing_extensions import Self


class Database(ValueEnum):
    """
    Database identifiers for provenance IDs
    """

    ICSD = "icsd"
    Pauling_Files = "pf"
    COD = "cod"


class Author(BaseModel):
    """
    Author information
    """

    name: str | None = Field(None)
    email: str | None = Field(None)


def _remove_dupe_authors(authors: list[dict[str, Any] | Author]):
    """Remove duplicate authors from a list of Author objects or their dict rep."""
    _authors = [a.model_dump() if hasattr(a, "model_dump") else a for a in authors]
    authors_dict = {a["name"].lower(): a for a in _authors}
    return list(authors_dict.values())


def _migrate_legacy_history_data(
    config: list[dict[str, Any]] | list[History],
) -> list[History]:
    """Migrate legacy provenance and SNL `history` data as a classmethod.

    Parameters
    -----------
    config : list of history data as a dict or History object

    Returns
    -----------
    list of History objects.
    """

    top_level_history = [
        h.model_dump() if isinstance(h, History) else h for h in config
    ]
    history: list[History] = []
    for inp_hist in top_level_history:
        descs: list = []
        _flatten_nested_descriptions(descs, inp_hist.get("description"), 0)
        history.extend(
            [
                History(
                    **{k: inp_hist.get(k) for k in ("name", "url")},  # type: ignore[arg-type]
                    description=desc,
                )
                for desc in descs
            ]
        )
    return history


def _flatten_nested_descriptions(
    descs: list[ProvenanceDescription | None],
    entry: dict[str, Any] | None,
    depth: int,
    remark: str | None = None,
) -> None:
    """Flatten legacy provenance description data.

    Parameters
    -----------
    descs : list[ProvenanceDescription | None]
        Running list of provenance description data.
    entry : dict[str,Any] or None
        The current description
    depth : int
        Used to indicate how far recursively into the history
        this field is nested.
    remark : str or None (default)
        Optional remark to supersede that in entry.

    Returns
    -----------
    None : all data goes in to `descs`
    """

    if not entry:
        descs.append(None)
        return

    elif nested_hist := entry.pop("history", []):
        for sub_entry in nested_hist:
            _flatten_nested_descriptions(
                descs, sub_entry, depth + 1, remark=f"Nested history depth {depth}"
            )

    elif init_args := entry.get("init_args"):
        _flatten_nested_descriptions(
            descs, init_args, depth + 1, remark=f"init_args-depth-{depth}"
        )

    # always append current entry even if there were nested fields
    orig_remark = entry.pop("remark", None)
    if not remark:
        remark = orig_remark

    if entry.get("lattice"):
        entry["lattice"] = Lattice.from_dict(entry["lattice"])

    for k in ("sites", "input_structure"):
        if c := entry.get(k):
            try:
                if all(c.get(x) for x in ("sites", "lattice")):
                    entry[k] = Structure.from_sites(
                        [
                            PeriodicSite.from_dict(
                                site, lattice=Lattice.from_dict(c["lattice"])
                            )
                            for site in c["sites"]
                        ]
                    )
                else:
                    entry[k] = Structure.from_sites(
                        [
                            PeriodicSite.from_dict(site, lattice=entry.get("lattice"))
                            for site in c
                        ]
                    )
            except Exception:
                entry[k] = None

    if isinstance(entry.get("species_map"), list):
        # Some of these appear to be .items()
        entry["species_map"] = {v[0]: v[1] for v in entry["species_map"]}

    if isinstance(cif_data := entry.get("cif_data"), dict):
        # CIF fields from pymatgen are too heterogeneous to schematize.
        entry["cif_data"] = json.dumps(cif_data)

    descs.append(ProvenanceDescription(**entry, remark=remark))


class TransformationMeta(BaseModel):
    """Schema for transformation fields in history.description."""

    transformation_name: str | None = None
    formula: str | None = None
    structureid: int | None = None


class ProvenanceDescription(BaseModel):
    """Schema for heterogeneous provenance description data."""

    monty_class: str | None = Field(None, validation_alias="@class")
    monty_module: str | None = Field(None, validation_alias="@module")
    IZA_code: str | None = None
    algo: int | None = None
    cif_data: str | None = None
    crystal_id: int | None = None
    datetime: str | None = None
    description: str | None = None
    disordered_crystal_id: int | None = None
    experimental: bool = False
    fraction_to_remove: float | None = None
    fw_id: int | None = None
    icsd_id: int | None = None
    id: str | None = None
    input_structure: StructureType | None = None
    lattice: LatticeType | None = None
    materialsid: IdentifierType | None = None
    name: str | None = None
    original_file: str | None = None
    oxidation_states: dict[str, int] | None = None
    query: str | None = None
    remark: str | None = None
    scaling_matrix: Matrix3D | None = None
    sites: StructureType | None = None
    source: str | None = None
    spacegroup: SymmetryData | None = None
    specie_to_remove: str | None = None
    species_map: dict[str, str] | None = None
    species_to_remove: list[str] | None = None
    string: str | None = None
    structureid: int | None = None
    task_id: str | None = None
    task_type: str | None = None
    transformation_name: str | None = None
    transformations: list[TransformationMeta] | None = None
    version: str | None = None

    @classmethod
    def parse_str_or_dict(cls, x: str | dict[str, Any] | Self) -> Self:
        """Method used in validating string or dict-like input.

        Parameters
        -----------
        x : a string, dict, or ProvenanceDescription

        Returns
        -----------
        ProvenanceDescription
        """
        if isinstance(x, str):
            return cls(string=x)  # type: ignore[arg-type,call-arg]
        elif isinstance(x, dict):
            return cls(**x)
        return x


class History(BaseModel):
    """
    History of the material provenance
    """

    name: str
    url: str
    description: Annotated[
        ProvenanceDescription | None,
        BeforeValidator(ProvenanceDescription.parse_str_or_dict),
    ] = Field(None, description="Dictionary of extra data for this history node.")


HistoryType = Annotated[
    list[History] | None,
    BeforeValidator(_migrate_legacy_history_data),
]


class SNLAbout(BaseModel):
    """A data dictionary defining extra fields in a SNL"""

    references: str = Field(
        "", description="Bibtex reference strings for this material."
    )

    authors: list[Author] | None = Field(
        None, description="list of authors for this material."
    )

    remarks: list[str] | None = Field(
        None, description="list of remarks for the provenance of this material."
    )

    tags: list[str] | None = Field(None)

    database_IDs: dict[Database, list[str]] | None = Field(
        None, description="Database IDs corresponding to this material."
    )

    history: HistoryType = Field(
        None,
        description="list of history nodes specifying the transformations or orignation"
        " of this material for the entry closest matching the material input.",
    )

    created_at: DateTimeType = Field(description="The creation date for this SNL.")

    @classmethod
    def migrate_legacy_data(cls, config: dict[str, Any]) -> Self:
        """Migrate legacy SNL data with free-form JSON values to schematized."""
        config["history"] = _migrate_legacy_history_data(config.get("history", []))
        return cls(**config)


class SNLDict(BaseModel):
    """Pydantic validated dictionary for SNL"""

    about: SNLAbout

    snl_id: str = Field(..., description="The SNL ID for this entry")


class ProvenanceDoc(PropertyDoc):
    """
    A provenance property block
    """

    property_name: str = "provenance"

    created_at: DateTimeType = Field(
        description="creation date for the first structure corresponding to this material",
    )

    references: list[str] = Field(
        default_factory=list, description="Bibtex reference strings for this material"
    )

    authors: Annotated[
        list[Author],
        BeforeValidator(_remove_dupe_authors),
    ] = Field(default_factory=list, description="list of authors for this material")

    remarks: list[str] | None = Field(
        None, description="list of remarks for the provenance of this material"
    )

    tags: list[str] | None = Field(None)

    theoretical: bool = Field(
        True, description="If this material has any experimental provenance or not"
    )

    database_IDs: dict[Database, list[str]] | None = Field(
        None, description="Database IDs corresponding to this material"
    )

    history: HistoryType = Field(
        default_factory=list,
        description="list of history nodes specifying the transformations or orignation"
        " of this material for the entry closest matching the material input",
    )

    @classmethod
    def from_SNLs(
        cls,
        structure: Structure,
        snls: list[SNLDict],
        material_id: IdentifierType | None = None,
        **kwargs,
    ) -> "ProvenanceDoc":
        """
        Converts legacy Pymatgen SNLs into a single provenance document
        """

        assert (
            len(snls) > 0
        ), "Error must provide a non-zero list of SNLs to convert from SNLs"

        # Choose earliest created_at
        created_at = min([snl.about.created_at for snl in snls])
        # last_updated = max([snl.about.created_at for snl in snls])

        # Choose earliest history
        history = sorted(snls, key=lambda snl: snl.about.created_at)[0].about.history

        # Aggregate all references into one dict to remove duplicates
        refs = {}
        for snl in snls:
            try:
                set_strict_mode(False)
                entries = parse_string(snl.about.references, bib_format="bibtex")
                refs.update(entries.entries)
            except Exception as e:
                warnings.warn(
                    f"Failed parsing bibtex: {snl.about.references} due to {e}"
                )

        bib_data = BibliographyData(entries=refs)

        references = [ref.to_string("bibtex") for ref in bib_data.entries.values()]

        # TODO: Maybe we should combine this robocrystallographer?
        # TODO: Refine these tags / remarks
        remarks = list(set([remark for snl in snls for remark in snl.about.remarks]))  # type: ignore[union-attr]
        tags = [r for r in remarks if len(r) < 140]

        authors = [entry for snl in snls for entry in snl.about.authors]  # type: ignore[union-attr]

        # Check if this entry is experimental
        exp_vals = []
        for snl in snls:
            for entry in snl.about.history:  # type: ignore[union-attr]
                if entry.description is not None:
                    exp_vals.append(entry.description.experimental)

        experimental = any(exp_vals)

        # Aggregate all the database IDs
        snl_ids = {snl.snl_id for snl in snls}
        db_ids = {
            Database(db_id): [snl_id for snl_id in snl_ids if db_id in snl_id]
            for db_id in map(str, Database)  # type: ignore
        }

        # remove Nones and empty lists
        db_ids = {k: list(filter(None, v)) for k, v in db_ids.items()}
        db_ids = {k: v for k, v in db_ids.items() if len(v) > 0}

        fields = {
            "created_at": created_at,
            "references": references,
            "authors": authors,
            "remarks": remarks,
            "tags": tags,
            "database_IDs": db_ids,
            "theoretical": not experimental,
            "history": history,
        }

        return super().from_structure(
            material_id=material_id, meta_structure=structure, **fields, **kwargs
        )

    @classmethod
    def migrate_legacy_data(cls, config: dict[str, Any]) -> Self:
        """Migrate legacy provenance data with free-form JSON values to schematized."""
        config["history"] = _migrate_legacy_history_data(config.get("history", []))
        return cls(**config)



"""Define equation of state (EOS) schemas."""

from __future__ import annotations

from pydantic import BaseModel, Field, BeforeValidator
from typing import TYPE_CHECKING, Annotated

from pymatgen.analysis.eos import EOS, EOSError

from emmet.core.material import BasePropertyMetadata
from emmet.core.types.enums import ValueEnum

if TYPE_CHECKING:
    from collections.abc import Sequence
    from typing import Any

    from pymatgen.core import Structure


class LegacyEOSModel(ValueEnum):
    """EOS models used to fit legacy data.

    Is overly permissive in allowing for
    spelling mistakes present in legacy data.
    """

    MIE_GRUNEISEN = "mie_gruneisen"
    PACK_EVANS_JAMES = "pack_evans_james"
    BIRCH = "birch"
    VINET = "vinet"
    TAIT = "tait"
    BIRCH_EULER = "birch_euler"
    POIRIER_TARANTOLA = "poirier_tarantola"
    BIRCH_LAGRANGE = "birch_lagrange"
    MURNAGHAN = "murnaghan"
    BIRCH_MURNAGHAN = "birch_murnaghan"

    @classmethod
    def _missing_(cls, value):
        if value == "pourier_tarantola":
            return cls.POIRIER_TARANTOLA


# Subset of EOS models known to pymatgen
PYMATGEN_KNOWN_EOS_MODELS = {
    LegacyEOSModel(eos_name)
    for eos_name in ("murnaghan", "birch", "birch_murnaghan", "vinet")
}


class EOSFit(BaseModel):
    """Schematize fitted EOS data.

    The only required field is `model`.
    If no other fields are instantiated, it is assumed
    that the fit failed.
    """

    model: LegacyEOSModel = Field(description="The EOS model used to fit the data.")
    V0: float | None = Field(None, description="The equilibrium volume in Å³.")
    B0: float | None = Field(None, description="The equilibrium bulk modulus, in GPa.")
    B1: float | None = Field(
        None,
        description="The pressure derivative of the bulk modulus at V0, dimensionless.",
    )
    E0: float | None = Field(None, description="The equilibrium energy, in eV.")
    R2: float | None = Field(
        None, description="The fit R2 (coefficient of determination), dimensionless."
    )

    @classmethod
    def from_ev_data(
        cls,
        volumes: Sequence[float],
        energies: Sequence[float],
        model: str | LegacyEOSModel,
    ):
        eos_pars: dict[str, float] = {}
        eos_model = LegacyEOSModel(model)
        try:
            eos = EOS(eos_name=eos_model.value).fit(volumes, energies)
            eos_pars.update({k.upper(): eos.results[k] for k in ("e0", "v0", "b1")})
            eos_pars["B0"] = float(eos.b0_GPa)

            sum_res_sq = ((eos.energies - eos.func(eos.volumes)) ** 2).sum()
            sum_tot_sq = ((eos.energies - eos.energies.mean()) ** 2).sum()
            eos_pars["R2"] = 1.0 - sum_res_sq / sum_tot_sq

        except EOSError:
            pass
        return cls(model=eos_model, **eos_pars)


def _migrate_legacy_data(
    config: dict[str, Any] | list[dict[str, Any]] | None,
) -> list[EOSFit] | None:
    """Migrate legacy data to new EOS schema."""
    migration = {"V0": "V0", "E0": "E0", "B": "B0", "C": "B1"}
    if isinstance(config, dict):
        config = [
            EOSFit(  # type: ignore[misc]
                **{v: config[model].get(k) for k, v in migration.items()},
                model=LegacyEOSModel(model),
            )
            for model in set(config).intersection({x.value for x in LegacyEOSModel})  # type: ignore[attr-defined]
        ]
    elif isinstance(config, list) and all(isinstance(x, dict) for x in config):
        config = [EOSFit(**x) for x in config]  # type: ignore[misc]

    return config  # type: ignore[return-value]


class EOSDoc(BasePropertyMetadata):
    """Fitted equations of state, and energy-volume fit data."""

    energies: list[float] | None = Field(
        None,
        description="Energies in eV that the equations of state are plotted with.",
    )

    volumes: list[float] | None = Field(
        None,
        description="Volumes in A³ that the equations of state are plotted with.",
    )

    eos: Annotated[list[EOSFit] | None, BeforeValidator(_migrate_legacy_data)] = Field(
        None,
        description="Data for each type of equation of state.",
    )

    @classmethod
    def from_ev_data(
        cls,
        structure: Structure,
        volumes: Sequence[float],
        energies: Sequence[float],
        models: list[str | LegacyEOSModel] | None = None,
    ):
        return cls.from_structure(
            meta_structure=structure,
            energies=energies,
            volumes=volumes,
            eos=[
                EOSFit.from_ev_data(volumes, energies, model)
                for model in (models or PYMATGEN_KNOWN_EOS_MODELS)  # type: ignore[union-attr]
            ],
        )



from pydantic import BaseModel, Field
from pymatgen.core.periodic_table import Element

from emmet.core.types.pymatgen_types.structure_adapter import MoleculeType


class MoleculesDoc(BaseModel):
    """
    Molecules relevant to battery electrolytes.
    """

    elements: list[Element] | None = Field(
        None,
        description="List of elements in the molecule.",
    )

    nelements: int | None = Field(
        None,
        description="Number of elements in the molecule.",
    )

    EA: float | None = Field(
        None,
        description="Electron affinity of the molecule in eV.",
    )

    IE: float | None = Field(
        None,
        description="Ionization energy of the molecule in eV.",
    )

    charge: int | None = Field(
        None,
        description="Charge of the molecule in +e.",
    )

    pointgroup: str | None = Field(
        None,
        description="Point group of the molecule in Schoenflies notation.",
    )

    smiles: str | None = Field(
        None,
        description="The simplified molecular input line-entry system (SMILES) \
            representation of the molecule.",
    )

    task_id: str | None = Field(
        None,
        description="Materials Project molecule ID. This takes the form mol-*****.",
    )

    molecule: MoleculeType | None = Field(
        None,
        description="Pymatgen molecule object.",
    )

    formula_pretty: str | None = Field(
        None,
        description="Chemical formula of the molecule.",
    )

    svg: str | None = Field(
        None,
        description="String representation of the SVG image of the molecule.",
    )



"""Annotate Bader charge/spin analysis output types from pymatgen."""

from pydantic import BaseModel


class BaderAnalysis(BaseModel):
    """Output of pymatgen.command_line.bader_caller.bader_analysis_from_objects

    We omit the `charge_densities` field, since these are too large
    to justify storing in the document model.

    Charge densities can already be stored in TaskDoc.vasp_objects
    """

    min_dist: list[float]
    charge: list[float]
    atomic_volume: list[float]
    vacuum_charge: float
    vacuum_volume: float
    reference_used: bool
    bader_version: float
    charge_transfer: list[float] | None = None
    magmom: list[float] | None = None



import pytest

# from matcalc.utils import get_universal_calculator
import numpy as np
from pymatgen.core import Structure
from pymatgen.util.testing import STRUCTURES_DIR

from emmet.core.elasticity import BulkModulus, ElasticTensorDoc, ShearModulus
from emmet.core.math import matrix_3x3_to_voigt
from emmet.core.ml import MLDoc, MatPESTrainDoc
from emmet.core.tasks import TaskDoc
from emmet.core.testing_utils import DataArchive

from tests.conftest import get_test_object


# if TYPE_CHECKING:
#    from ase.calculators.calculator import Calculator

expected_keys = {
    # -- metadata --
    "material_id": str,
    "structure": Structure,
    "deprecated": bool,
    "matcalc_version": type(None),  # str,
    "model_name": type(None),  # str,
    "model_version": type(None),  # str,
    # -- relaxation --
    "final_structure": Structure,
    "energy": float,
    "volume": float,
    "a": float,
    "b": float,
    "c": float,
    "alpha": float,
    "beta": float,
    "gamma": float,
    # -- eos --
    "eos": dict,
    "bulk_modulus_bm": float,
    # -- phonon --
    "temperatures": list,
    "free_energy": list,
    "entropy": list,
    "heat_capacity": list,
    # -- elasticity --
    "elastic_tensor": ElasticTensorDoc,
    "shear_modulus": ShearModulus,
    "bulk_modulus": BulkModulus,
    "young_modulus": float,
}


# @pytest.mark.parametrize(
#    ("calculator", "prop_kwargs"),
#    [
#        (get_universal_calculator("chgnet"), None),
#        ("M3GNet-MP-2021.2.8-PES", {"ElasticityCalc": {"relax_structure": False}}),
#    ],
# )
@pytest.mark.skip(reason="Temporary skip. Needs attention.")
def test_ml_doc(calculator, prop_kwargs: dict) -> None:
    struct = Structure.from_file(STRUCTURES_DIR / "Si.json")

    doc = MLDoc(
        structure=struct,
        calculator=calculator,
        material_id="mp-33",
        deprecated=False,
        prop_kwargs=prop_kwargs,
    )

    # check that all expected keys are present
    missing = sorted({*expected_keys} - {*doc.__dict__})
    assert not missing, f"keys {missing=}"

    # check that all keys have expected type
    for key, typ in expected_keys.items():
        actual = getattr(doc, key)
        assert isinstance(
            actual, typ
        ), f"{key=} expected type={typ.__name__}, got {type(actual).__name__}"


def test_matpes_doc_from_task_doc(test_dir):

    with DataArchive.extract(
        test_dir / "vasp" / f"{get_test_object('SiOptimizeDouble').folder}.json.gz"
    ) as dir_name:
        task_doc = TaskDoc.from_directory(dir_name)

    matpes_train_docs = MatPESTrainDoc.from_task_doc(task_doc)

    assert len(matpes_train_docs) == sum(
        len(cr.output.ionic_steps) for cr in task_doc.calcs_reversed
    )

    ctr = 0
    for cr in task_doc.calcs_reversed[::-1]:
        for iistep, istep in enumerate(cr.output.ionic_steps):
            assert matpes_train_docs[ctr].energy == pytest.approx(istep.e_0_energy)
            assert np.allclose(matpes_train_docs[ctr].forces, istep.forces)

            assert np.allclose(
                matpes_train_docs[ctr].stress, matrix_3x3_to_voigt(istep.stress)
            )

            if iistep < len(cr.output.ionic_steps) - 1:
                assert matpes_train_docs[ctr].bandgap is None
                assert (
                    matpes_train_docs[ctr].structure.site_properties.get("magmom")
                    is None
                )
            else:
                assert matpes_train_docs[ctr].bandgap == pytest.approx(
                    cr.output.bandgap
                )
                assert np.allclose(
                    matpes_train_docs[ctr].structure.site_properties.get("magmom"),
                    cr.output.structure.site_properties.get("magmom"),
                )

            ctr += 1



"""Legacy MP tools to validate VASP calculations."""

from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np
from pydantic import BaseModel, Field, ImportString
from pymatgen.core.structure import Structure
from pymatgen.io.vasp.inputs import Kpoints
from pymatgen.io.vasp.sets import VaspInputSet

from emmet.core.base import EmmetBaseModel
from emmet.core.settings import EmmetSettings
from emmet.core.tasks import CoreTaskDoc, TaskDoc
from emmet.core.types.enums import DeprecationMessage
from emmet.core.types.typing import DateTimeType, IdentifierType
from emmet.core.vasp.calc_types.enums import CalcType, TaskType
from emmet.core.vasp.calculation import CoreCalculationOutput
from emmet.core.vasp.task_valid import TaskDocument

if TYPE_CHECKING:
    from collections.abc import Sequence

SETTINGS = EmmetSettings()


class ValidationDataDict(BaseModel):
    encut_ratio: float | None = Field(None)
    max_gradient: float | None = Field(None)
    kspacing_delta: float | None = Field(None)
    kpts_ratio: float | None = Field(None)


class ValidationDoc(EmmetBaseModel, extra="allow"):
    """
    Validation document for a VASP calculation
    """

    task_id: IdentifierType = Field(
        ..., description="The task_id for this validation document"
    )
    valid: bool = Field(False, description="Whether this task is valid or not")
    last_updated: DateTimeType = Field(
        description="Last updated date for this document",
    )
    reasons: list[DeprecationMessage | str] | None = Field(
        None, description="List of deprecation tags detailing why this task isn't valid"
    )
    warnings: list[str] = Field(
        [], description="List of potential warnings about this calculation"
    )
    data: ValidationDataDict | None = Field(
        None,
        description="Dictioary of data used to perform validation."
        " Useful for post-mortem analysis",
    )
    nelements: int | None = Field(None, description="Number of elements.")
    symmetry_number: int | None = Field(
        None,
        title="Space Group Number",
        description="The spacegroup number for the lattice.",
    )
    chemsys: str | None = Field(None)
    formula_pretty: str | None = Field(None)

    @classmethod
    def from_task_doc(
        cls,
        task_doc: CoreTaskDoc | TaskDoc | TaskDocument,
        kpts_tolerance: float = SETTINGS.VASP_KPTS_TOLERANCE,
        kspacing_tolerance: float = SETTINGS.VASP_KSPACING_TOLERANCE,
        input_sets: dict[str, ImportString] = SETTINGS.VASP_DEFAULT_INPUT_SETS,
        LDAU_fields: list[str] = SETTINGS.VASP_CHECKED_LDAU_FIELDS,
        max_allowed_scf_gradient: float = SETTINGS.VASP_MAX_SCF_GRADIENT,
        max_magmoms: dict[str, float] = SETTINGS.VASP_MAX_MAGMOM,
        potcar_stats: dict[CalcType, dict[str, str]] | None = None,
    ) -> "ValidationDoc":
        """
        Determines if a calculation is valid based on expected input parameters from a pymatgen inputset

        Args:
            task_doc: the task document to process
            kpts_tolerance: the tolerance to allow kpts to lag behind the input set settings
            kspacing_tolerance:  the tolerance to allow kspacing to lag behind the input set settings
            input_sets: a dictionary of task_types -> pymatgen input set for validation
            pseudo_dir: directory of pseudopotential directory to ensure correct hashes
            LDAU_fields: LDAU fields to check for consistency
            max_allowed_scf_gradient: maximum uphill gradient allowed for SCF steps after the
                initial equillibriation period
            potcar_stats: Dictionary of potcar stat data. Mapping is calculation type -> potcar symbol -> hash value.
        """

        nelements = task_doc.nelements or None
        symmetry_number = task_doc.symmetry.number if task_doc.symmetry else None

        bandgap = task_doc.output.bandgap
        calc_type = task_doc.calc_type
        task_type = task_doc.task_type
        run_type = task_doc.run_type
        chemsys = task_doc.chemsys
        formula_pretty = task_doc.formula_pretty

        if isinstance(task_doc, (TaskDoc, TaskDocument)):
            inputs = task_doc.orig_inputs
            calcs_reversed = [
                calc if not hasattr(calc, "model_dump") else calc.model_dump()
                for calc in task_doc.calcs_reversed
            ]

            if calcs_reversed[0].get("input", {}).get("structure", None):
                structure = calcs_reversed[0]["input"]["structure"]
            else:
                structure = task_doc.input.structure or task_doc.output.structure
        else:
            inputs = task_doc.input
            calcs_reversed = None
            structure = task_doc.input.structure or task_doc.output.structure

        if isinstance(structure, dict):
            structure = Structure.from_dict(structure)

        reasons = []
        data = {}  # type: ignore
        warnings: list[str] = []

        if str(calc_type) in input_sets:
            try:
                valid_input_set = _get_input_set(
                    run_type, task_type, calc_type, structure, input_sets, bandgap
                )

            except (TypeError, KeyError, ValueError):
                reasons.append(DeprecationMessage.SET)
                valid_input_set = None

            try:
                # Sometimes spglib can't determine space group with the default
                # `symprec` and `angle_tolerance`. In these cases,
                # `Structure.get_space_group_info()` fails
                valid_input_set.structure.get_space_group_info()
            except Exception:
                reasons.append(DeprecationMessage.SYMMETRY)
                valid_input_set = None

            if valid_input_set:
                # Checking POTCAR summary_stats if a directory is supplied
                if potcar_stats:
                    if _potcar_stats_check(task_doc, potcar_stats):
                        if task_type in [
                            TaskType.NSCF_Line,
                            TaskType.NSCF_Uniform,
                            TaskType.DFPT_Dielectric,
                            TaskType.Dielectric,
                        ]:
                            warnings.append(DeprecationMessage.POTCAR.__doc__)  # type: ignore
                        else:
                            reasons.append(DeprecationMessage.POTCAR)

                # Checking K-Points
                # Calculations that use KSPACING will not have a .kpoints attr

                if task_type != TaskType.NSCF_Line:
                    # Not validating k-point data for line-mode calculations as constructing
                    # the k-path is too costly for the builder and the uniform input set is used.

                    if valid_input_set.kpoints is not None:
                        if _kpoint_check(
                            valid_input_set,
                            inputs,
                            calcs_reversed,
                            data,
                            kpts_tolerance,
                        ):
                            reasons.append(DeprecationMessage.KPTS)

                    else:
                        # warnings
                        _kspacing_warnings(
                            valid_input_set, inputs, data, warnings, kspacing_tolerance
                        )

                # warn, but don't invalidate if wrong ISMEAR
                valid_ismear = valid_input_set.incar.get("ISMEAR", 1)
                incar = inputs.get("incar", {})
                curr_ismear = incar.get("ISMEAR", 1)
                if curr_ismear != valid_ismear:
                    warnings.append(
                        f"Inappropriate smearing settings. Set to {curr_ismear},"
                        f" but should be {valid_ismear}"
                    )

                # Checking ENCUT
                encut = incar.get("ENCUT")
                valid_encut = valid_input_set.incar["ENCUT"]
                data["encut_ratio"] = float(encut) / valid_encut  # type: ignore
                if data["encut_ratio"] < 1:
                    reasons.append(DeprecationMessage.ENCUT)

                # U-value checks
                if _u_value_checks(task_doc, valid_input_set, warnings):
                    reasons.append(DeprecationMessage.LDAU)

                if calcs_reversed:
                    # Check the max upwards SCF step
                    if _scf_upward_check(
                        calcs_reversed, inputs, data, max_allowed_scf_gradient, warnings
                    ):
                        reasons.append(DeprecationMessage.MAX_SCF)

                # Check for Am and Po elements. These currently do not have proper elemental entries
                # and will not get treated properly by the thermo builder.
                if ("Am" in chemsys) or ("Po" in chemsys):
                    reasons.append(DeprecationMessage.MANUAL)

                # Check for magmom anomalies for specific elements
                if _magmom_check(
                    calcs_reversed or task_doc.output,
                    structure,
                    max_magmoms=max_magmoms,
                ):
                    reasons.append(DeprecationMessage.MAG)

            else:
                if "Unrecognized" in str(calc_type):
                    reasons.append(DeprecationMessage.UNKNOWN)
                else:
                    reasons.append(DeprecationMessage.SET)

        doc = ValidationDoc(
            task_id=task_doc.task_id,
            calc_type=calc_type,
            run_type=task_doc.run_type,
            valid=len(reasons) == 0,
            reasons=reasons,
            data=data,
            warnings=warnings,
            nelements=nelements,
            symmetry_number=symmetry_number,
            chemsys=chemsys,
            formula_pretty=formula_pretty,
        )

        return doc


def _get_input_set(run_type, task_type, calc_type, structure, input_sets, bandgap):
    # Ensure inputsets get proper additional input values
    if "SCAN" in run_type.value:
        valid_input_set: VaspInputSet = input_sets[str(calc_type)](structure, bandgap=bandgap)  # type: ignore
    elif task_type == TaskType.NSCF_Uniform or task_type == TaskType.NSCF_Line:
        # Constructing the k-path for line-mode calculations is too costly, so
        # the uniform input set is used instead and k-points are not checked.
        valid_input_set = input_sets[str(calc_type)](structure, mode="uniform")

    elif task_type == TaskType.NMR_Electric_Field_Gradient:
        valid_input_set = input_sets[str(calc_type)](structure, mode="efg")

    else:
        valid_input_set = input_sets[str(calc_type)](structure)

    return valid_input_set


def _scf_upward_check(calcs_reversed, inputs, data, max_allowed_scf_gradient, warnings):
    skip = abs(inputs.get("incar", {}).get("NELMDL", -5)) - 1
    energies = [
        d["e_fr_energy"]
        for d in calcs_reversed[0]["output"]["ionic_steps"][-1]["electronic_steps"]
    ]
    if len(energies) > skip:
        max_gradient = np.max(np.gradient(energies)[skip:])
        data["max_gradient"] = max_gradient
        if max_gradient > max_allowed_scf_gradient:
            return True
    else:
        warnings.append(
            "Not enough electronic steps to compute valid gradient"
            " and compare with max SCF gradient tolerance"
        )
        return False


def _u_value_checks(task_doc, valid_input_set, warnings):
    # NOTE: Reverting to old method of just using input.hubbards which is wrong in many instances
    input_hubbards = {} if task_doc.input.hubbards is None else task_doc.input.hubbards

    if valid_input_set.incar.get("LDAU", False) or len(input_hubbards) > 0:
        # Assemble required input_set LDAU params into dictionary
        input_set_hubbards = dict(
            zip(
                valid_input_set.poscar.site_symbols,
                valid_input_set.incar.get("LDAUU", []),
            )
        )

        all_elements = list(set(input_set_hubbards.keys()) | set(input_hubbards.keys()))
        diff_ldau_params = {
            el: (input_set_hubbards.get(el, 0), input_hubbards.get(el, 0))
            for el in all_elements
            if not np.allclose(input_set_hubbards.get(el, 0), input_hubbards.get(el, 0))
        }

        if len(diff_ldau_params) > 0:
            warnings.extend(
                [
                    f"U-value for {el} should be {good} but was {bad}"
                    for el, (good, bad) in diff_ldau_params.items()
                ]
            )
            return True

    return False


def _kpoint_check(input_set, inputs, calcs_reversed, data, kpts_tolerance):
    """
    Checks to make sure the total number of kpoints is correct
    """
    valid_num_kpts = input_set.kpoints.num_kpts or np.prod(input_set.kpoints.kpts[0])

    if calcs_reversed:
        input_dict = calcs_reversed[0].get("input", {})

        if not input_dict:
            input_dict = inputs

    else:
        input_dict = inputs

    kpoints = input_dict.get("kpoints", {})
    if isinstance(kpoints, Kpoints):
        kpoints = kpoints.as_dict()
    elif kpoints is None:
        kpoints = {}
    num_kpts = kpoints.get("nkpoints", 0) or np.prod(kpoints.get("kpoints", [1, 1, 1]))

    data["kpts_ratio"] = num_kpts / valid_num_kpts
    return data["kpts_ratio"] < kpts_tolerance


def _kspacing_warnings(input_set, inputs, data, warnings, kspacing_tolerance):
    """
    Issues warnings based on KSPACING values
    """
    valid_kspacing = input_set.incar.get("KSPACING", 0)
    if kspacing := inputs.get("incar", {}).get("KSPACING"):
        data["kspacing_delta"] = kspacing - valid_kspacing
        # larger KSPACING means fewer k-points
        if data["kspacing_delta"] > kspacing_tolerance:
            warnings.append(
                f"KSPACING is greater than input set: {data['kspacing_delta']}"
                f" lower than {kspacing_tolerance} "
            )
        elif data["kspacing_delta"] < kspacing_tolerance:
            warnings.append(
                f"KSPACING is lower than input set: {data['kspacing_delta']}"
                f" lower than {kspacing_tolerance} "
            )


def _potcar_stats_check(
    task_doc,
    potcar_stats: dict,
    exclude_keys: Sequence[str] | None = ["sha256", "copyr"],
):
    """
    Checks to make sure the POTCAR summary stats is equal to the correct
    value from the pymatgen input set.
    """
    data_tol = 1.0e-6
    excl: set[str] = set([k.lower() for k in (exclude_keys or [])])

    try:
        potcar_details = task_doc.calcs_reversed[0].model_dump()["input"]["potcar_spec"]
    except AttributeError:
        potcar_details = task_doc.input.model_dump()["potcar_spec"]
    except KeyError:
        # Assume it is an old calculation without potcar_spec data and treat it as passing POTCAR hash check
        return False

    use_legacy_hash_check = False
    if any(entry.get("summary_stats", None) is None for entry in potcar_details):
        # potcar_spec doesn't include summary_stats kwarg needed to check potcars
        # fall back to header hash checking
        use_legacy_hash_check = True

    all_match = True
    for entry in potcar_details:
        if not entry["titel"]:
            all_match = False
            break

        symbol = entry["titel"].split(" ")[1]
        ref_summ_stats = potcar_stats[str(task_doc.calc_type)].get(symbol, None)

        if not ref_summ_stats:
            # Symbol differs from reference set - deprecate
            all_match = False
            break

        if use_legacy_hash_check:
            all_match = any(
                all(
                    entry[key] == ref_stat[key]
                    for key in (
                        "hash",
                        "titel",
                    )
                )
                for ref_stat in ref_summ_stats
            )

        else:
            entry_keys = {
                key: set([k.lower() for k in entry["summary_stats"]["keywords"][key]])
                - excl
                for key in ["header", "data"]
            }
            all_match = False
            for ref_stat in ref_summ_stats:
                ref_keys = {
                    key: set([k.lower() for k in ref_stat["keywords"][key]]) - excl
                    for key in ["header", "data"]
                }

                key_match = all(entry_keys[k] == v for k, v in ref_keys.items())

                data_match = False
                if key_match:
                    data_match = all(
                        abs(
                            ref_stat["stats"][key][stat]
                            - entry["summary_stats"]["stats"][key][stat]
                        )
                        < data_tol
                        for stat in ["MEAN", "ABSMEAN", "VAR", "MIN", "MAX"]
                        for key in ["header", "data"]
                    )
                all_match = key_match and data_match

                if all_match:
                    # Found at least one match to reference POTCAR summary stats,
                    # that suffices for the check
                    break

        if not all_match:
            break

    return not all_match


def _magmom_check(
    calc: list | CoreCalculationOutput,
    structure: Structure,
    max_magmoms: dict[str, float],
):
    """
    Checks for maximum magnetization values for specific elements.
    Returns True if the maximum absolute value outlined below is exceded for the associated element.
    """

    if isinstance(calc, CoreCalculationOutput):
        outcar = calc.outcar
    else:
        outcar = calc[0]["output"]["outcar"]

    if outcar and (mag_info := outcar.get("magnetization", [])):
        return any(
            abs(mag_info[isite].get("tot", 0.0))
            > abs(max_magmoms.get(site.label, np.inf))
            for isite, site in enumerate(structure)
        )

    return False


def _get_unsorted_symbol_set(structure: Structure):
    """
    Have to build structure_symbol set manually to ensure
    we get the right order since pymatgen sorts its symbol_set list.
    """
    return list(
        {
            str(sp): 1 for site in structure for sp, v in site.species.items() if v != 0
        }.keys()
    )



"""Common models and types needed for VASP data."""

from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np
from pydantic import BaseModel, Field

from pymatgen.core import Structure
from pymatgen.io.common import VolumetricData as PmgVolumetricData

from emmet.core.types.enums import ValueEnum
from emmet.core.types.pymatgen_types.structure_adapter import StructureType

if TYPE_CHECKING:
    from collections.abc import Callable
    from typing_extensions import Self


class ElectronicStep(BaseModel):
    """Document defining the information at each electronic step.

    Note, not all the information will be available at every step.
    """

    alphaZ: float | None = Field(None, description="The alpha Z term.")
    ewald: float | None = Field(None, description="The ewald energy.")
    hartreedc: float | None = Field(None, description="Negative Hartree energy.")
    XCdc: float | None = Field(None, description="Negative exchange energy.")
    pawpsdc: float | None = Field(
        None, description="Negative potential energy with exchange-correlation energy."
    )
    pawaedc: float | None = Field(None, description="The PAW double counting term.")
    eentropy: float | None = Field(None, description="The entropy (T * S).")
    bandstr: float | None = Field(
        None, description="The band energy (from eigenvalues)."
    )
    atom: float | None = Field(None, description="The atomic energy.")
    e_fr_energy: float | None = Field(None, description="The free energy.")
    e_wo_entrp: float | None = Field(None, description="The energy without entropy.")
    e_0_energy: float | None = Field(None, description="The internal energy.")


class VolumetricLabel(ValueEnum):
    TOTAL = "total"
    DIFF = "diff"

    # NB: we only need these if we have noncolinear calculations
    # DIFF_X = "diff_x"
    # DIFF_Y = "diff_y"
    # DIFF_Z = "diff_z"


class AugChargeData(BaseModel):
    label: str | None = Field(
        None, description="The label written by VASP for the augmenation charge data."
    )
    data: list[float] | None = Field(None, description="The augmentation charges.")


class ChgcarLike(BaseModel):
    """Model for VASP CHGCAR-like data.

    Includes CHGCAR, CHG, LOCPOT, ELFCAR, AECCAR0, AECCAR1, AECCAR2, POT.
    All volumetric data is flattened upon storage in the model from a
    rank-3 tensor.
    C-order is used throughout.
    """

    labels: list[VolumetricLabel] | None = Field(
        None, description="The spin resolution of the volumetric data."
    )
    data: list[list[float] | None] | None = Field(
        description="The primary volumetric data."
    )
    data_rank: list[tuple[int, int, int]] | None = Field(
        None, description="The original shape of the"
    )
    data_aug: list[list[AugChargeData] | None] | None = Field(
        None, description="The augmentation charge volumetric data."
    )
    structure: StructureType | None = Field(
        None, description="The structure associated with the volumetric data."
    )

    @staticmethod
    def parse_augmentation_charge_data(
        aug_data: dict[str, list[str]],
    ) -> dict[VolumetricLabel, list[AugChargeData]]:
        aug_data_arr: dict[VolumetricLabel, list[AugChargeData]] = {}
        for k, unfmt_data in aug_data.items():
            if not any(line.strip() for line in unfmt_data):
                continue
            parse_meta = True
            num_vals = -1
            aug_data_arr[VolumetricLabel(k)] = []
            atom_data: list[float] = []
            atom_label: str | None = None
            for row in unfmt_data:
                data = row.replace("\n", "").split()
                if parse_meta:
                    if not data[0].isalpha():
                        # pymatgen sometimes puts extra lines here because they
                        # exist in a CHGCAR but have no clear meaning.
                        # probably needs a fix in pymatgen
                        continue

                    label = " ".join([x for x in data[:-1] if x.isalpha()])

                    atom_label = label
                    atom_data = []
                    num_vals = int(data[-1])
                    parse_meta = False
                else:
                    atom_data.extend([float(x) for x in data])
                    if len(atom_data) >= num_vals:
                        parse_meta = True
                        aug_data_arr[VolumetricLabel(k)].append(
                            AugChargeData(label=atom_label, data=atom_data)
                        )

        return aug_data_arr

    @classmethod
    def from_pmg(cls, vd: PmgVolumetricData) -> Self:
        """Convert generic pymatgen volumetric data to an archive format."""
        labels = [VolumetricLabel(k) for k in vd.data]
        data_aug = None
        if aug_data := cls.parse_augmentation_charge_data(vd.data_aug):  # type: ignore[arg-type]
            data_aug = [aug_data.get(vlab) for vlab in labels]

        return cls(
            labels=labels,
            data=[vd.data[vlab].flatten(order="C") for vlab in labels],  # type: ignore[misc]
            data_rank=[vd.data[vlab].shape for vlab in labels],  # type: ignore[misc]
            data_aug=data_aug,
            structure=(
                vd.structure
                if isinstance(vd.structure, Structure)
                else Structure.from_dict(vd.structure.as_dict())
            ),
        )

    def to_pmg(self, pmg_cls: Callable = PmgVolumetricData) -> PmgVolumetricData:
        """Convert to a pymatgen VolumetricData-derived object."""
        data_aug: dict[str, np.ndarray] | None = None
        if self.data_aug:
            data_aug = {
                vol_label.value: np.array(self.data_aug[i])
                for i, vol_label in enumerate(self.labels or [])
                if self.data_aug[i]
            }
        return pmg_cls(
            self.structure,
            {
                vol_label.value: np.array(self.data[i]).reshape(  # type: ignore[index]
                    self.data_rank[i], order="C"  # type: ignore[index]
                )
                for i, vol_label in enumerate(self.labels or [])
            },
            data_aug=data_aug,
        )
